package clustering.suffixtree;

import java.util.*;

public class InternalNode extends Node {

    ArrayList children; //Set?

    //New internal node with given label
    public InternalNode(String[] label, int offset, int length) {
        super(label, offset, length);
        children = new ArrayList();
        indices = new HashSet();
    }

    //Creates node from nofInCommon last tokens of existing node and copies children and indices
    public InternalNode(InternalNode old, int nofInCommon) {
        super(old.label, old.labelOffset + nofInCommon, (old.labelLength - nofInCommon));
        children = old.children;
        indices = old.indices;
    }

    //Creates new InternalNode from an existing Leafnode (copies indices)
    public InternalNode(LeafNode old) {
        super(old.label, old.labelOffset, old.labelLength);
        indices = old.indices;
        children = new ArrayList();
    }


    public void addString(String[] label, int offset, Integer index) {
        Node child;
        if ((child = hasChild(label[offset])) != null) {
            //Fant barn som matcher frste ord
            int nofInCommon = nofInCommon(child, label, offset);

            if (nofInCommon == child.labelLength) {
                //Matcher hele eksisterende label
                if (child.labelLength == (label.length - offset)) {
                    //Helt like? Legg til instance
                    child.indices.add(index);
                } else {
                    //Label = nofInCommon, m lage ny internal med insts og ny lv under
                    if (child instanceof LeafNode) {
                        //M lage ny intern og legge p rest av label som nytt barn
                        InternalNode newInternal = new InternalNode((LeafNode) child);
                        LeafNode newLeaf = new LeafNode(label, offset + nofInCommon); // flytt
                        newLeaf.indices.add(index);
                        newInternal.children.add(newLeaf);
                        this.children.remove(child); //TODO: engine...
                        this.children.add(newInternal);
                    } else {
                        //Child er Intern, lage nytt barn der...
                        child.addString(label, offset + nofInCommon, index);
                    }

                }
            } else {
                //Partly matching -> split the tree

                //Create new InternalNode with the common characters (labelOffset -> labelOffset + nofInCommon)
                InternalNode newInternal = new InternalNode(label, offset, nofInCommon); //No instances...
                
                //Create restNode for rest of the original child (copy indices and possibly children)
                Node newChildOldLabel;
                if (child instanceof LeafNode)
                    //Lag ny lv med child's occs og label - labelOffset frste
                    newChildOldLabel = new LeafNode((LeafNode) child, child.labelOffset + nofInCommon);
                else
                    //Lag ny intern med child's barn, indices og label = oldLabel - nofInCommon frste
                    newChildOldLabel = new InternalNode((InternalNode) child, nofInCommon);
                newInternal.children.add(newChildOldLabel);

                if (nofInCommon == (label.length - offset)) {
                    //newInternalNode is as long as the new label, just add the indices
                    newInternal.indices.add(index);
                } else {
                    //new label is longer than nofInCommon, create new leaf for this part.
                    //Create restNode for the new label (labelOffset + nofInCommon -> resten)
                    LeafNode newChildNewLabel = new LeafNode(label, offset + nofInCommon, label.length - offset - nofInCommon, index);
                    newInternal.addChild(newChildNewLabel);
                }

                //Remove existing child and add the new child
                children.remove(child);
                children.add(newInternal);
            }
        } else {
            //Fant ingen -> lag nytt barn
            children.add(new LeafNode(label, offset, label.length - offset, index));
        }
    }

    //Sker efter pattern[labelOffset->] og legger treff i indices
    public void search(String[] pattern, int offset, Set indices) {
        Node child;
        if ((child = hasChild(pattern[offset])) != null) {
            //Fant barn som matcher frste bokstav.
            int nofInCommon = nofInCommon(child, pattern, offset);

            if (nofInCommon == (pattern.length - offset)) {
                //Labelet inneholder patternet, returner indices under denne noden
                if (child instanceof InternalNode) {
                    getIndicesBeneath((InternalNode) child, indices);
                }
                indices.addAll(child.indices);
            } else if (nofInCommon == child.labelLength) {
                //Hele labelet matcher -> Vi m ske videre
                ((InternalNode) child).search(pattern, offset + nofInCommon, indices);
            }
        }
    }

    public static void getIndicesBeneath(InternalNode internalNode, Set indices) {
        ArrayList children = internalNode.children;
        for (int i = 0; i < children.size(); i++) {
            Node child = (Node) children.get(i);
            indices.addAll(child.indices);
            if (child instanceof InternalNode) {
                //Go deeper;
                getIndicesBeneath((InternalNode) child, indices);
            }
        }
    }

    //Har node (fra labelOffset til labelOffset+labelLength) noe felles med label fra labelOffset og ut?)
    private int nofInCommon(Node node, String[] label, int offset) {
        int nof = Math.min(node.labelLength, label.length - offset);
        for (int i = 0; i < nof; i++) {
            if (!node.label[node.labelOffset + i].equals(label[offset + i]))
                return i;
        }
        return nof;
    }

    public void addChild(Node node) {
        children.add(node);
    }

    public Node hasChild(String word) {
        for (int i = 0; i < children.size(); i++) {
            Node child = (Node) children.get(i);
            if (child.label[child.labelOffset].equals(word))
                return child;
        }
        return null;
    }
}
