package clustering.suffixtree;

import tools.StopwordRemover;
import tools.PorterStemmer;

import java.util.Map.Entry;
import java.util.*;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;

import searching.Settings;
import clustering.stc.SuffixTreeClusterer;
import clustering.stc.Cluster;
import clustering.stc.BaseCluster;

public class SuffixTree {

    Map wordToDocIDSet = new HashMap(); //Maps each word to a set of Document IDs
    int docCounter = 1;
    InternalNode root = new InternalNode(new String[]{"moder"}, 0, 1);
    StopwordRemover stopwordRemover = StopwordRemover.getInstance();
    PorterStemmer stemmer = PorterStemmer.getInstance();
    Map labelToOriginalString = new HashMap();

    public static void main(String[] args) {
        long a = System.currentTimeMillis();
        SuffixTree st = new SuffixTree();
        /*st.addSnippet("... Jaguar car reviews  Review Centre ");
        st.addSnippet("Wild jaguars - PANTERA ONCA");
        st.addSnippet("JAGUAR - Jaguar car reviews!");
        st.addSnippet("... pantera onca pictures");*/
        //st.addSnippet("... pantera onca picture ...");

        st.addSnippet("Jaguar car reviews  Review Centre", new Integer(1));
        st.addSnippet("## PANTERA ONCA ##", new Integer(2));
        st.addSnippet("Jaguar reviews!", new Integer(3));
        st.addSnippet("Buy Pantera Onca Pictures", new Integer(4));

        /*st.addSnippet("## PANTERA ONCA ##");
        st.addSnippet("... pantera onca picture ...");
        st.addSnippet("Buy Jaguar Pictures");
        st.addSnippet("Jaguar cars");*/


        SuffixTreeClusterer clusterer = new SuffixTreeClusterer(st, new HashMap(), new HashSet());
        Cluster[] clusters = clusterer.clusterDocs();
        System.out.println("Fant " + clusters.length + " clusters p " + (System.currentTimeMillis() - a) + "ms");
        for (int i = 0; i < clusters.length; i++) {
            Cluster cluster = clusters[i];
            System.out.println("[" + i + "] '" + cluster.label + "' (" + cluster.getSize() + ")");
        }

    }

    public void addSnippet(String snippet, Integer docID) {
        docCounter++;
        try {
            Reader in = new StringReader(snippet + "\n"); //Just in case...
            int val;
            char ch;
            ArrayList sentence = new ArrayList();
            char[] buffer = new char[200]; //MaxWordLength (snippets are only 150 chars..)
            int bufferPtr = 0;
            int currentPos = 0;//Holds position in this snippet
            while ((val = in.read()) != -1) {
                ch = (char) val;
                if ((val >= 65 && val <= 90) || (val >= 97 && val <= 122)) {
                    //Found part of a word (A-Z || a-z)
                    buffer[bufferPtr++] = ch;
                } else {
                    //SpecialChar -> New word
                    boolean startNewSentence = false;

                    //Add word if its longer than 2 chars
                    if (bufferPtr > 2) {
                        //Strip last char if word ends with an -s and second
                        //last letter is not s (business, press etc)
                        String newWord = new String(buffer, 0, bufferPtr).toLowerCase();
                        if (Settings.REMOVE_SNIPPET_STOPWORDS && stopwordRemover.isStopword(newWord))
                            startNewSentence = true;
                        else {
                            sentence.add(newWord);
                        }
                    }

                    //Start new buffer if it contains any chars
                    if (bufferPtr > 0) {
                        buffer = new char[255];
                        bufferPtr = 0;
                    }

                    //Start new sentence if not space
                    startNewSentence |= val != 32;

                    if (startNewSentence) {
                        //Add new sentence if it contains any words
                        if (sentence.size() > 0) {

                            //Preprocess the words, map the processed string to the original string:
                            String[] originalSentence = (String[]) sentence.toArray(new String[sentence.size()]);
                            String[] processedSentence = preprocessSentence(originalSentence);
                            addSentence(processedSentence, docID);
                        }
                        sentence = new ArrayList();
                    }
                }
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private String[] preprocessSentence(String[] originalSentence) {
        String[] processedSentence = new String[originalSentence.length];
        for (int i = 0; i < originalSentence.length; i++) {
            String word = originalSentence[i];

            if (Settings.STEM_SNIPPETS)
                word = stemmer.getStemmedWord(word);
            else {
                //Just remove -s if second last char is not s (e.g. business)
                int length = word.length();
                if (word.charAt(length-1) == 's' && word.charAt(length-2) != 's')
                    word = word.substring(0, length-1);
            }

            processedSentence[i] = word;
            labelToOriginalString.put(processedSentence, originalSentence);
        }
        return processedSentence;
    }

    //Used for looking up the original string when displaying cluster labels
    public String getOriginalLabel(BaseCluster bc) {
        String[] label = (String[]) labelToOriginalString.get(bc.labelString);
        StringBuffer sb = new StringBuffer(label[bc.originOffset]);
        for (int i = bc.originOffset + 1; i < (bc.originOffset + bc.originLength); i++) {
            sb.append(" " + label[i]);
        }
        return sb.toString();
    }

    public void addSentence(String string) {
        addSentence(string.split(" "), new Integer(docCounter++));
    }

    private void addSentence(String[] tokens, Integer docID) {
        //Index the words:
        for (int i = 0; i < tokens.length; i++) {
            Set set = (Set) wordToDocIDSet.get(tokens[i]);
            if (set == null) {
                set = new HashSet();
                wordToDocIDSet.put(tokens[i], set);
            }
            set.add(docID);
        }

        //Add all suffixes to the tree
        for (int i = 0; i < tokens.length; i++) {
            root.addString(tokens, i, docID);
        }
    }

    public ArrayList getBaseClusters() {
        Map map = new HashMap(); //Maps strings to docIDSet
        ArrayList children = ((InternalNode) root).children;
        for (int i = 0; i < children.size(); i++) {
            Node child = (Node) children.get(i);
            getBaseClusterAndIndices(child, map, "", 0);
        }

        ArrayList nodes = new ArrayList();
        Set entries = map.entrySet();
        for (Iterator q = entries.iterator(); q.hasNext();) {
            Entry entry = (Entry) q.next();
            nodes.add(new BaseCluster((String) entry.getKey(), (Object[]) entry.getValue()));
        }

        return nodes;
    }

    /**
     * @param root         Gathers indices from this node
     * @param map          Map to organize docIDs for each phrase
     * @param currentLabel label of root.parentNode
     * @return set of indices found in/under the root
     */
    private Set getBaseClusterAndIndices(Node root, Map map, String currentLabel, int currentDepth) {
        Set indices = new HashSet();
        indices.addAll(root.indices); //Add indices of currently processed node
        if (root instanceof LeafNode && root.indices.size() == 1) {
            //Not a basecluster, just return DocID
        } else {
            //BaseCluster. Process children if it's an internal node.
            String label = (currentLabel.length() > 0) ? (currentLabel + " " + root.getLabel()) : (root.getLabel());
            if (root instanceof InternalNode) {
                InternalNode newRoot = (InternalNode) root;
                for (int i = 0; i < newRoot.children.size(); i++) {
                    Node newChild = (Node) newRoot.children.get(i);
                    Set deeperIndices = getBaseClusterAndIndices(newChild, map, label, currentDepth + root.labelLength);
                    indices.addAll(deeperIndices);
                }
            }

            //Add label to map and addAll indices found here and underneath
            Object[] obj = (Object[]) map.get(label);
            if (obj == null)
                //Map new label to indices and the Node (used to retrieve original string info)
                map.put(label, new Object[] {indices, root, new Integer(currentDepth)});
            else {
                //Add all indices to the Set that label maps to
                Set existingSet = (Set) obj[0];
                existingSet.addAll(indices);
            }
        }
        return indices;
    }

    public int nofDocs() {
        return docCounter;
    }

    public Map getWordToDocIDSet() {
        return wordToDocIDSet;
    }

    //Returns list of indices
    public Set search(String pattern) {
        HashSet indices = new HashSet();
        long a = System.currentTimeMillis();
        root.search(pattern.toLowerCase().split(" "), 0, indices);
        long b = System.currentTimeMillis();
        System.out.println("Found " + indices.size() + " matches in " + (b - a) + "ms.");
        return indices;
    }
}
