package crawling.odp;

import tools.StopwordRemover;
import tools.PorterStemmer;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;

import searching.Settings;
import searching.Settings;

public class ODPCategoryMiner {

    private static StopwordRemover stopwordRemover = StopwordRemover.getInstance();
    private static PorterStemmer stemmer = PorterStemmer.getInstance();

    public static void main(String[] args) throws Exception {
        WordWeightPair[] wordsAndWeights = indexWords();
        writeHashMapToDisk(wordsAndWeights);

        /*for (int i = 0; i < 100; i++) {
            System.out.println(wordsAndWeights[i].word + ":\t\t(" + wordsAndWeights[i].tf + "\t\t" + wordsAndWeights[i].idf + "\t\t" + wordsAndWeights[i].avg + "\t" + wordsAndWeights[i].nof + ") = " + wordsAndWeights[i].weight);
        }
        for (int i = wordsAndWeights.length - 100; i < wordsAndWeights.length; i++) {
            System.out.println(wordsAndWeights[i].word + ":\t\t(" + wordsAndWeights[i].tf + "\t\t" + wordsAndWeights[i].idf + "\t\t" + wordsAndWeights[i].avg + "\t" + wordsAndWeights[i].nof + ") = " + wordsAndWeights[i].weight);
        } */
        //getCategoriesFromStructureFile();
    }

    private static void writeHashMapToDisk(WordWeightPair[] wordsAndWeights) throws Exception {
        ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream("indices/odpcategories.dat"));
        Map map = new HashMap(wordsAndWeights.length);
        for (int i = 0; i < wordsAndWeights.length; i++) {
            WordWeightPair pair = wordsAndWeights[i];
            map.put(pair.word, new Float(pair.weight));
        }
        oos.writeObject(map);
        oos.close();
    }

    public static WordWeightPair[] indexWords() throws Exception {
        Map map = new HashMap();
        String inn = "odp/categories.reduced.txt";
        BufferedReader br = new BufferedReader(new FileReader(inn));
        String line;
        int maxFreq = 0;
        int nofCategories = 0;

        //Create Map<String,ArrayList<Integer>> mapping word to list of depths
        while ((line = br.readLine()) != null) {
            nofCategories++;
            int depth = getDepth(line);
            String[] tokens = getTokens(line);
            for (int i = 0; i < tokens.length; i++) {
                String token = tokens[i];
                ArrayList list = (ArrayList) map.get(token);
                if (list == null) {
                    list = new ArrayList();
                    map.put(token, list);
                }
                list.add(new Integer(depth));
                //Keep track of highest frequency of any word
                if (list.size() > maxFreq)
                    maxFreq = list.size();
            }
        }

        System.out.println("Nof words " + map.size() + " highest freq " + maxFreq);

        Set entries = map.entrySet();
        WordWeightPair[] wordsAndWeights = new WordWeightPair[entries.size()];
        int i = 0;
        for (Iterator iterator = entries.iterator(); iterator.hasNext();) {
            Entry entry = (Entry) iterator.next();
            String word = (String) entry.getKey();
            ArrayList depths = (ArrayList) entry.getValue();

            //Get average depth
            float averageDepth = 0;
            for (int j = 0; j < depths.size(); j++) {
                Integer integer = (Integer) depths.get(j);
                averageDepth += integer.intValue();
            }
            averageDepth /= (float) depths.size();

            //Calculate TFxIDF
            //float weight = (depths.size() / (float) maxFreq) * (float) Math.log(nofCategories / (float) depths.size()) * averageDepth;
            //wordsAndWeights[i] = new WordWeightPair(word, weight);
            //wordsAndWeights[i] = new WordWeightPair(word, ((depths.size() + 10) / (float) maxFreq), (float) Math.log(nofCategories / (float) depths.size()), averageDepth, depths.size());
            wordsAndWeights[i] = new WordWeightPair(word, ((depths.size() + 10) / (float) maxFreq), (float) Math.log(nofCategories / (float) depths.size()), averageDepth, depths.size());

            i++;
        }
        Arrays.sort(wordsAndWeights);
        br.close();
        return wordsAndWeights;
    }

    static class WordWeightPair implements Comparable {

        String word;
        float weight;
        float tf;
        float idf;
        float avg;
        int nof;

        public WordWeightPair(String word, float tf, float idf, float avg, int nof) {
            this.word = word;
            this.tf = tf;
            this.idf = idf;
            this.avg = avg;
            this.weight = (float) Math.sqrt(1 + tf * idf * avg);
            this.nof = nof;
        }

        public WordWeightPair(String word, float weight) {
            this.word = word;
            this.weight = weight;
        }

        public int compareTo(Object o) {
            return (int)(1000*(((WordWeightPair)o).weight - weight));
        }

    }

    public static int getDepth(String ODPLabel) {
        char[] chars = ODPLabel.toCharArray();
        int depth = 0;
        for (int i = 0; i < chars.length; i++) {
            if (chars[i] == '/')
                depth++;
        }
        return depth;
    }

    public static String[] getTokens(String FullODPLabel) {
        int pos = FullODPLabel.lastIndexOf('/');
        if (pos == -1)
            return new String[] {};
        String ODPLabel = FullODPLabel.substring(pos);
        ODPLabel = ODPLabel.replace('_', ' ');
        String[] tokens = ODPLabel.split(" ");
        ArrayList cleanTokens = new ArrayList();
        for (int i = 0; i < tokens.length; i++) {
            String token = tokens[i];
            if (token.length() > 1) {
                char[] chars = token.toCharArray();
                boolean isLegal = true;
                for (int j = 0; j < chars.length; j++) {
                    int val = (int) chars[j];
                    if (!(val >= 65 && val <= 90) && !(val >= 97 && val <= 122)) {
                        isLegal = false;
                        break;
                    }
                }
                token = token.toLowerCase();
                if (isLegal && !stopwordRemover.isStopword(token) && !token.equals("america")) {
                    if (Settings.STEM_ODP_LABELS)
                        token = stemmer.getStemmedWord(token);
                    cleanTokens.add(token);
                }
            }
        }
        return (String[]) cleanTokens.toArray(new String[cleanTokens.size()]);
    }

    public static void getCategoriesFromStructureFile() {
        try {
            String inn = "odp/structure.rdf.u8";
            String ut = "odp/categories2.txt";
            BufferedReader br = new BufferedReader(new FileReader(inn));
            BufferedWriter bw = new BufferedWriter(new FileWriter(ut));
            String line = br.readLine();
            while (line != null) {

                //Read until next category
                while (line != null && !line.startsWith("<Topic"))
                    line = br.readLine();

                if (line == null)
                    break;

                String category = line.substring(13, line.lastIndexOf('\"'));

                if (!category.startsWith("Top/Regional") && !category.startsWith("Top/World") && !category.startsWith("Top/Adult")) {
                    bw.write(category + "\n");

                    //Read until end of topic, save description
                    /*String description = null;
                    while (!line.startsWith("</Topic")) {
                        line = br.readLine();
                        if (line.startsWith("  <d:Description")) {
                            description = line.substring(17)
                        }
                    }
                    bw.write(line.substring(17) + "\n");*/
                }

                line = br.readLine();
            }
            br.close();
            bw.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
