package searching.dataobjects;

import java.io.*;

/**
 * Manages the document collection. Reads document info from docinfo.dat.
 * Retrieves document info and generates snippets.
 */
public class DocumentManager {

    private Document[] documents;
    private int nofDocs;
    private short[] highestFreq; //Stores the frequency for most frequent word in document i
    private float[] vectorLengths; //Sqrt of sum of squared weights (see cosine similarity)
    private String documentFolder; //Stores the basefolder to which the filenames are relative

    public DocumentManager(String docInfoFilename) {
        readDocInfoFromFile(docInfoFilename);
    }

    public Document getDocument(int docID) {
        return documents[docID];
    }

    public int getNofDocs() {
        return nofDocs;
    }

    private void readDocInfoFromFile(String docInfoFilename) {
        long start = System.currentTimeMillis();
        ObjectInputStream docInfoFile;
        try {
            docInfoFile = new ObjectInputStream(new FileInputStream(docInfoFilename));
            nofDocs = docInfoFile.readInt();
            highestFreq = (short[]) docInfoFile.readObject();
            vectorLengths = (float[]) docInfoFile.readObject();
            documentFolder = (String) docInfoFile.readObject();
            documents = (Document[]) docInfoFile.readObject();
            docInfoFile.close();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
        //System.out.println("Read document info in " + (System.currentTimeMillis() - start) + "ms.");
    }

    public float getVectorLength(int docID) {
        return vectorLengths[docID];
    }

    public short getHighestFreq(int docID) {
        return highestFreq[docID];
    }

    /**
     * Returns a snippet of 150 chars from filepointer onwards
     */
    public String getSnippet(int docID, long filepointer) {
        try {
            //TODO: if filepointer is near end of file, it should skip backwards... (file.len - filepointer < 150)
            Reader in = new FileReader(documentFolder + "/" + documents[docID].filename);
            char[] buffer = new char[150];
            in.skip(filepointer); //Skip ahead to the filepointer
            int length = in.read(buffer);
            String snippet = new String(buffer, 0, length);
            if (filepointer != 0)
                snippet = "..." + snippet;
            if (length == 150)
                snippet += "...";
            return snippet;
        } catch (IOException e) {
            e.printStackTrace();
            return null;
        }
    }
}
