package searching.engine;

import tools.StopWatch;
import searching.dataobjects.DocumentManager;
import searching.dataobjects.Result;
import searching.dataobjects.Document;

/**
 * Class that allows searching an engine written to disk by searching.engine.IndexWriter
 */
public class IndexSearcher {

    private Index index;
    private DocumentManager docManager;

    public IndexSearcher(String indexFolder) {
        index = new Index(indexFolder);
        docManager = index.getDocumentManager();
    }

    /**
     * Searches for query and generates snippets for the first nofSnippets results
     */
    public Result[] search(String query, int nofSnippets) {

        //Search
        StopWatch t = new StopWatch();
        Result[] results = index.query(query);
        searchTime = t.reset();

        //Assign Documents and generate snippets
        for (int i = 0; i < results.length; i++) {
            Result result = results[i];
            Document doc = docManager.getDocument(result.docID);
            if (i < nofSnippets)
                doc.snippet = docManager.getSnippet(result.docID, result.filepointer);
            result.document = doc;
        }
        snippetGenerationTime = t.reset();

        return results;
    }

    private long snippetGenerationTime = 0;
    private long searchTime = 0;

    /**
     * Method for testing purposes, gets times for last search operation
     */
    public int[] getSearchAndSnippetGenerationTime() {
        return new int[] {(int) searchTime, (int) snippetGenerationTime};
    }

    /**
     * Gets filename for a given documentID
     */
    public String getFileName(int docID) {
        return docManager.getDocument(docID).filename;
    }
}
