package web;

import org.springframework.web.servlet.mvc.AbstractController;
import org.springframework.web.servlet.ModelAndView;

import javax.servlet.http.*;


import java.util.*;
import java.io.*;

import tools.StopWatch;
import clustering.stc.SuffixTreeClusterer;
import clustering.stc.Cluster;
import clustering.stc.BaseCluster;
import searching.engine.IndexSearcher;
import searching.dataobjects.Result;

public class MainController {

    static IndexSearcher searcher;
    static Map ODPLabelWeights;
    static Set queryLog;
    static int nofDocsToCluster;
    static int nofClustersDisplayed = 10;
    static int nofDocsPerPage = 30;

    static PrintStream out;

    public static void runPerformanceTests(String[] args) throws Exception {
        MainController mc = new MainController();
        //Prepp query and search
        String actualQuery = args[0];
        nofDocsToCluster = Integer.parseInt(args[2]);
        int nof = Integer.parseInt(args[1]);
        //System.out.println("Clustering " + actualQuery + " with " + nofDocsToCluster + " " + nof + " times");
        Result[] results = searcher.search(actualQuery, nofDocsToCluster);
        //System.out.println("nofResults = " + results.length);
        //System.out.println("nof\tsuffix\tcluster\ttotal");

        //Create list of snippets and cluster the docs:
        int actualNofClusterDocs = Math.min(nofDocsToCluster, results.length);
        ArrayList snippetList = new ArrayList(actualNofClusterDocs);
        ArrayList docIDList = new ArrayList(actualNofClusterDocs);
        for (int i = 0; i < actualNofClusterDocs; i++) {
            snippetList.add(results[i].document.snippet);
            docIDList.add(new Integer(results[i].document.docID));
        }
        StopWatch t;
        SuffixTreeClusterer stc;
        Cluster[] clusters;
        long clusteringMs, suffixMs;
        long totalClustering = 0, totalSuffix = 0;
        float[] values = new float[nof];
        for (int i = 0; i < nof; i++) {
            t = new StopWatch();
            stc = new SuffixTreeClusterer(snippetList, docIDList, ODPLabelWeights, queryLog);
            suffixMs = t.reset();
            clusters = stc.clusterDocs();
            clusteringMs = t.time();
            totalClustering += clusteringMs;
            totalSuffix += suffixMs;
            //System.out.println("Clustering " + nofDocsToCluster + " took " + clusteringMs + "ms");
        }

        float avgClustering = (float) totalClustering / (float) nof;
        float avgSuffix = (float) totalSuffix / (float) nof;
        float avg = (int)avgClustering + (int)avgSuffix;
        System.out.println(nofDocsToCluster + "\t" + (int)avgClustering + "\t" + (int)avgSuffix + "\t" + (int)avg);

        /*float sd = 0;
        for (int i = 0; i < nof; i++) {
            sd += Math.pow((values[i] - avg), 2);
        }
        sd = (float) Math.sqrt(sd / (float) nof);

        System.out.println("Standard deviation: " + sd + "ms");*/
        //System.out.println("Found " + clusters.length + " clusters");
        //calculateClusterSims(clusters);
    }

    public static void main(String[] args) throws Exception {

        boolean runPerformanceTests = false;
        if (runPerformanceTests) {
            runPerformanceTests(args);
            System.exit(0);
        }

        MainController mc = new MainController();
        //Prepp query and search
        String actualQuery = "paris";
        nofDocsToCluster = 200;
        Result[] results = searcher.search(actualQuery, nofDocsToCluster);

        //Create list of snippets and cluster the docs:
        int actualNofClusterDocs = Math.min(nofDocsToCluster, results.length);
        ArrayList snippetList = new ArrayList(actualNofClusterDocs);
        ArrayList docIDList = new ArrayList(actualNofClusterDocs);
        for (int i = 0; i < actualNofClusterDocs; i++) {
            snippetList.add(results[i].document.snippet);
            docIDList.add(new Integer(results[i].document.docID));
        }
        StopWatch t = new StopWatch();
        SuffixTreeClusterer stc = new SuffixTreeClusterer(snippetList, docIDList, ODPLabelWeights, queryLog);
        Cluster[] clusters = stc.clusterDocs();
        long clusteringMs = t.time();
        System.out.println("Clustering " + nofDocsToCluster + " took " + clusteringMs + "ms");
        System.out.println("Found " + clusters.length + " clusters");
        //calculateClusterSims(clusters);

        Set files = new HashSet();

        for (int i = 0; i < 10; i++) {
            Cluster cluster = clusters[i];
            BaseCluster[] BCs = (BaseCluster[]) cluster.baseClusters.toArray(new BaseCluster[cluster.baseClusters.size()]);
            Arrays.sort(BCs);

            //Select top 10 and scramble (if more than two elements)
            /*if (BCs.length > 2) {
                int len = Math.min(BCs.length, 10);
                Set newSet = new HashSet();
                for (int j = 0; j < len; j++) {
                    newSet.add(BCs[j]);
                }
                BaseCluster[] newBCs = new BaseCluster[len];
                int pos = 0;
                for (Iterator iterator = newSet.iterator(); iterator.hasNext();) {
                    newBCs[pos++] = (BaseCluster) iterator.next();
                }
                BCs = newBCs;
            } */

            //Print for latex: (husk  ikke kjr scramble)
            /*System.out.print("\\noindent\\textbf{" + cluster.getLabel() + " (" + cluster.getSize() + "):} \\texttt{\\{" + BCs[0].label);
            for (int j = 0; j < BCs.length; j++) {
                BaseCluster bc = BCs[j];
                //System.out.println(bc.label + "\t" + bc.b_length + "\t" + bc.f_of_p_length + "\t" + bc.score  + "\t" + bc.coverage + "\t" + bc.odpboost  + "\t" + bc.queryboost);
                //System.out.println(bc.label + "\t" + (int)(bc.coverage*100) + "%");
                System.out.print(stc.getLabelForBaseCluster(bc) + " (" + (int)(bc.coverage*100) + "%), ");
            }
            System.out.println("\\}}\\\\");*/

            //Printscreen
            System.out.print((i+1) + " & \\{" + stc.getLabelForBaseCluster(BCs[0]));
            for (int j = 1; j < BCs.length; j++) {
                BaseCluster bc = BCs[j];
                //System.out.println(bc.label + "\t" + bc.b_length + "\t" + bc.f_of_p_length + "\t" + bc.score  + "\t" + bc.coverage + "\t" + bc.odpboost  + "\t" + bc.queryboost);
                //System.out.println(bc.label + "\t" + (int)(bc.coverage*100) + "%");
                System.out.print(", " + stc.getLabelForBaseCluster(bc));
            }
            System.out.println("\\} \\\\");

            //Print for excel

            /*//Get first 5 titles
            Set docIDs = cluster.getDocIDs();
            int nof = 0;
            String[] titles = new String[6];
            for (int j = 0; j < results.length; j++) {
                Result result = results[j];
                if (docIDs.contains(new Integer(result.docID))) {
                    titles[nof++] = result.getTitle();
                    if (nof == 6)
                        break;
                }
            }

            System.out.print("Cluster " + (i+1) + " (" + cluster.getSize() + " documents):\t\t" + cluster.getLabel() + "\t\t");

            for (int j = 0; j < titles.length; j++) {
                System.out.print(titles[j] + "");
            }

            System.out.println("");


            for (int j = 0; j < BCs.length; j++) {
                BaseCluster bc = BCs[j];
                System.out.println(stc.getLabelForBaseCluster(bc) + "\t" + (int)(bc.coverage*100) + "%");
            }
            System.out.println("");*/

            //Print for KE:
            /*System.out.println(cluster.getLabel() + " (" + cluster.getSize() + " documents):");
                        BufferedWriter bw = new BufferedWriter(new FileWriter("collections/clusters/" + actualQuery + "." + i + ".txt"));
            Set ids = cluster.getDocIDs();
            for (Iterator iterator = ids.iterator(); iterator.hasNext();) {
                Integer docID = (Integer) iterator.next();
                String filename = searcher.getFileName(docID.intValue());
                try {
                    BufferedReader br = new BufferedReader(new FileReader("collections/queriesTagged/" + filename));
                    String line;
                    while ((line = br.readLine()) != null)
                        bw.write(line + "\n");
                    br.close();
                } catch (Exception e) {
                    System.out.println("couldn't open " + "collections/queriesTagged/" + filename);
                }
            }
            bw.close();*/
        }
        //out.close();
    }

    private static void calculateClusterSims(Cluster[] clusters) {
        float[][] sims = new float[clusters.length][clusters.length];
        for (int i = 0; i < clusters.length; i++) {
            for (int j = 0; j < clusters.length; j++) {
                sims[j][i] = overlap(clusters[j].getDocIDs(), clusters[i].getDocIDs());
            }
        }
        for (int i = 0; i < sims.length; i++) {
            for (int j = 0; j < sims.length; j++) {
                if (sims[i][j] > 0.5f)
                    System.out.print(sims[i][j] + " (" + clusters[i].getDocIDs().size() + "/" + clusters[j].getDocIDs().size() + ")\t");
                else
                    System.out.print("\t");
            }
            System.out.println("");
        }
    }

    //Hvor mange prosent av A er i B?
    private static float overlap(Set a, Set b) {

        //Get intersectionSize:
        int nof = 0;
        int aSize = a.size();
        int bSize = b.size();
        for (Iterator iterator = a.iterator(); iterator.hasNext();) {
            if (b.contains(iterator.next()))
                nof++;
        }

        return (float) nof / (float) aSize;
    }


    public MainController() throws Exception {
        nofDocsToCluster = 100;
        String base = "D:/data/skole/diplom/yahoo/";
        searcher = new IndexSearcher(base + "indices/queries");
        loadODPLabelWeights(base + "indices/odpcategories.dat");
        loadQueryLog(base + "indices/querylog.dat");
    }

    private static void loadODPLabelWeights(String filename) throws Exception {
        ObjectInputStream ois = new ObjectInputStream(new FileInputStream(filename));
        ODPLabelWeights = (Map) ois.readObject();
        ois.close();
    }

    private static void loadQueryLog(String filename) throws Exception {
        ObjectInputStream ois = new ObjectInputStream(new FileInputStream(filename));
        queryLog = (Set) ois.readObject();
        ois.close();
    }

    public ModelAndView handleRequestInternal(HttpServletRequest httpServletRequest, HttpServletResponse httpServletResponse) throws Exception {
        String query = httpServletRequest.getParameter("query");
        if (query == null || query.equals("null"))
            query = "";
        String nofClusterDocs = httpServletRequest.getParameter("nofDocsToCluster");
        if (nofClusterDocs == null || nofClusterDocs.equals("null"))
            nofClusterDocs = null;
        String filter = httpServletRequest.getParameter("filter");
        if (filter == null || filter.equals("null"))
            filter = null;

        HttpSession session = httpServletRequest.getSession(true);
        Map previousModel = (Map) session.getAttribute("model");
        if (filter != null && previousModel != null && previousModel.get("query").equals(query)) {
            //This is just a filtering operation, reuse old map:

            //Get the documents from the cluster:
            Result[] results = (Result[]) previousModel.get("allResults");
            Cluster[] clusters = (Cluster[]) previousModel.get("allClusters");
            int clusterID = Integer.parseInt(filter);
            int id;
            for (id = 0; id < clusters.length; id++) {
                if (clusters[id].getId() == clusterID)
                    break; //Just use i, no safety if you change filter manually
            }
            Cluster cluster = clusters[id];
            Set docIDs = cluster.getDocIDs();
            System.out.println("Filtering '" + query + "' on cluster '" + cluster.label + "'");

            //Get max 30 results from this cluster
            ArrayList filteredResultList = new ArrayList();
            for (int i = 0; i < results.length; i++) {
                Result result = results[i];
                if (docIDs.contains(new Integer(result.document.docID))) {
                    filteredResultList.add(result);
                    if (filteredResultList.size() == nofDocsPerPage)
                        break;
                }
            }
            Result[] filteredResults = (Result[]) filteredResultList.toArray(new Result[filteredResultList.size()]);
            previousModel.put("results", filteredResults);

            return new ModelAndView("enginefront", "model", previousModel);
        } else {
            //This is a normal query:
            if (query.equals(""))
                return new ModelAndView("largeenginefront");
            else {
                //Prepp query and search
                String actualQuery = query.trim();
                Result[] results = searcher.search(actualQuery, nofDocsToCluster);
                int nofResultsDisplayed = Math.min(results.length, nofDocsPerPage);
                int nofResults = results.length;
                int[] times = ((IndexSearcher) searcher).getSearchAndSnippetGenerationTime();
                float searchTime = (float) times[0] / 1000.0f;
                float snippetTime = (float) times[1] / 1000.0f;

                //Create list of snippets and cluster the docs:
                int actualNofClusterDocs = Math.min(nofDocsToCluster, results.length);
                ArrayList snippetList = new ArrayList(actualNofClusterDocs);
                ArrayList docIDList = new ArrayList(actualNofClusterDocs);
                for (int i = 0; i < actualNofClusterDocs; i++) {
                    snippetList.add(results[i].document.snippet);
                    docIDList.add(new Integer(results[i].document.docID));
                }
                StopWatch t = new StopWatch();
                SuffixTreeClusterer stc = new SuffixTreeClusterer(snippetList, docIDList, ODPLabelWeights, queryLog);
                Cluster[] clusters = stc.clusterDocs();
                long clusteringMs = t.time();
                float clusterTime = (float) clusteringMs / 1000.0f;

                //Trim clusterList?
                Cluster[] allClusters = clusters;
                if (clusters.length > nofClustersDisplayed) {
                    Cluster[] newClusters = new Cluster[nofClustersDisplayed];
                    System.arraycopy(clusters, 0, newClusters, 0, nofClustersDisplayed);
                    clusters = newClusters;
                }

                //Trim resultList?
                Result[] allResults = results;
                if (results.length > nofDocsPerPage) {
                    Result[] newResults = new Result[nofDocsPerPage];
                    System.arraycopy(results, 0, newResults, 0, nofDocsPerPage);
                    results = newResults;
                }

                //Store only results within the nofDocsToCluster (with snippets)
                if (allResults.length > nofDocsToCluster) {
                    Result[] newAllResults = new Result[nofDocsToCluster];
                    System.arraycopy(allResults, 0, newAllResults, 0, nofDocsToCluster);
                    allResults = newAllResults;
                }

                //Create model with the information gathered:
                Map model = new HashMap();
                model.put("query", query);
                model.put("searchTime", new Float(searchTime));
                model.put("snippetTime", new Float(snippetTime));
                model.put("clusterTime", new Float(clusterTime));
                model.put("results", results);
                model.put("allResults", allResults);
                model.put("nofResults", new Integer(nofResults));
                model.put("nofResultsDisplayed", new Integer(nofResultsDisplayed));
                model.put("clusters", clusters);
                model.put("allClusters", allClusters);
                session.setAttribute("model", model);
                return new ModelAndView("enginefront", "model", model);
            }
        }
    }
}
